!
!  artificial frames
!
!  Copyright © 2016-7 F.Hroch (hroch@physics.muni.cz)
!
!  This file is part of Munipack.
!
!  Munipack is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  Munipack is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with Munipack.  If not, see <http://www.gnu.org/licenses/>.
!
!
!  * The variable star part, implemented as a class, would by more classy.
!

program artificial

  use fitsio
  use trajd
  use generator
  use lcurves
  use lcio
  use iso_fortran_env

  implicit none

  integer, parameter :: dbl = selected_real_kind(15)
  real(dbl), parameter :: j2000 = 2451544.5_dbl

  character(len=4*FLEN_FILENAME) :: record,key,val,fmt
  character(len=FLEN_FILENAME) :: cat = '', phsystable = '', phsystem = 'Johnson', &
       file = 'artificial.fits', filemask = 'artificial_???.fits'
  logical :: verbose = .false., plog = .false.
  integer, dimension(2) :: naxes = (/ 666, 666 /)
  real(dbl), dimension(2) :: crpix = (/333, 333/)
  integer :: bitpix = -32
  integer :: nfiles = 0
  integer :: eq,n,i,l1,l2
  real(dbl) :: blevel = 1000, bnoise = -1
  real(dbl), dimension(2) :: bgrad = (/0, 0/)
  integer :: maglim = 13
  real(dbl) :: exptime = 1, area = 1     ! 1 sec per 1 sq.meter, equivalent r=0.56m
  real(dbl) :: hwhm = 1                  ! in pixels
  real(dbl) :: airy = 0.1                ! Airy disk radius in pixels for r=0.56m
  real(dbl) :: beta = 2                  ! Moffat exponent
  real(dbl) :: scale = 1.0 / 3600.0      ! 1 arcsex = 1 pix
  real(dbl) :: fov = 666. / 3600.0       ! field of view, used for scale setup only
  real(dbl) :: angle = 0
  real(dbl) :: qefficiency = 1
  real(dbl) :: extink = 0                ! extinction coefficient
  real(dbl) :: longitude = 0             ! geographic longitude
  real(dbl) :: latitude = 0              ! geographic latitude
  real(dbl) :: timestep = 60             ! time step in seconds
  real(dbl) :: jd = j2000                ! 2000-01-01T00:00:00
  real(dbl) :: year = 2000, month = 1, day = 1, hour = 0, min = 0, sec = 0, r, d
  real(dbl), dimension(2) :: crval = 0
  real(dbl) :: lcmag = 15                ! lc magnitude
  real(dbl) :: lcamp = 0.4               ! lc amplitude
  real(dbl) :: mamp = 1                  ! lc amplitude in magnitudes
  real(dbl) :: lcjd0 = j2000             ! lc reference time
  real(dbl) :: lcper = 1                 ! lc period
  real(dbl), dimension(2) :: lcoo = 0    ! lc coordinates
  character(len=FLEN_VALUE) :: filter = 'V'
  character(len=FLEN_VALUE), dimension(5) :: labels = (/'','','','',''/)
  character(len=FLEN_VALUE) :: psf = 'SEEING'
  character(len=FLEN_VALUE) :: spread = 'FFT'
  character(len=FLEN_VALUE) :: lcmodel = '', lctable = '', lcfourier = ''
  complex(dbl), dimension(:), allocatable :: fcoeff ! lc Fourier coefficients
  real(dbl), dimension(:), allocatable :: xcoo,ycoo,flux
  ! 'FFT' for FFT convolution, 'RANDOM' for random offset, 'AUTO' for 'RANDOM' of short
  ! exposures and 'TEST' for direct convolution
  logical :: atmosphere = .false.
  logical :: lcmode = .false.
  logical :: datetime_init = .false.
  logical :: lccoo_init = .false.

  do
     read(*,'(a)',end=20) record

     eq = index(record,'=')
     if( eq == 0 ) stop 'Improper input.'
     key = record(:eq-1)
     val = record(eq+1:)

     if( key == 'VERBOSE' ) then

        read(val,*) verbose

     else if( key == 'PIPELOG' ) then

        read(val,*) plog

     else if( key == 'PSF' ) then

        read(val,*) psf

     else if( key == 'SPREAD' ) then

        read(val,*) spread

     else if( key == 'NAXES' ) then

        read(val,*) naxes

     else if( key == 'BITPIX' ) then

        read(val,*) bitpix

     else if( key == 'BLEVEL' ) then

        read(val,*) blevel

     else if( key == 'BNOISE' ) then

        read(val,*) bnoise

     else if( key == 'BGRADX' ) then

        read(val,*) bgrad(1)

     else if( key == 'BGRADY' ) then

        read(val,*) bgrad(2)

     else if( key == 'MAGLIM' ) then

        read(val,*) maglim

     else if( key == 'HWHM' ) then

        read(val,*) hwhm

     else if( key == 'AIRY' ) then

        read(val,*) airy

     else if( key == 'BETA' ) then

        read(val,*) beta

     else if( key == 'FILTER' ) then

        read(val,*) filter

     else if( key == 'EXPTIME' ) then

        read(val,*) exptime

     else if( key == 'AREA' ) then

        read(val,*) area

     else if( key == 'QEFF' ) then

        read(val,*) qefficiency

     else if( key == 'EXTINK' ) then

        read(val,*) extink

     else if( key == 'FOV' ) then

        read(val,*) fov
        scale = fov / naxes(2)

     else if( key == 'DIAMETER' ) then

        read(val,*) d
        r = d / 2
        area = 3.14*r**2
        airy = 7e-2/r

     else if( key == 'ANGLE' ) then

        read(val,*) angle

     else if( key == 'SCALE' ) then

        read(val,*) scale

     else if( key == 'CRVAL' ) then

        read(val,*) crval

     else if( key == 'CAT' ) then

        read(val,*) cat

     else if( key == 'COL_RA' ) then

        read(val,*) labels(1)

     else if( key == 'COL_DEC' ) then

        read(val,*) labels(3)

     else if( key == 'COL_PMRA' ) then

        read(val,*) labels(2)

     else if( key == 'COL_PMDEC' ) then

        read(val,*) labels(4)

     else if( key == 'COL_MAG' ) then

        read(val,*) labels(5)

     else if( key == 'PHSYSTABLE' ) then

        read(val,*) phsystable

     else if( key == 'PHOTSYS' ) then

        read(val,*) phsystem

     else if( key == 'GEOGRAPHIC' ) then

        read(val,*) longitude, latitude

     else if( key == 'TIMESTEP' ) then

        read(val,*) timestep

     else if( key == 'DATE' ) then

        read(val,*) year,month,day
        datetime_init = .true.

     else if( key == 'TIME' ) then

        read(val,*) hour,min,sec
        datetime_init = .true.

     else if( key == 'ATMOSPHERE' ) then

        read(val,*) atmosphere

     else if( key == 'LCMODEL' ) then

        read(val,*) lcmodel

     else if( key == 'LCTABLE' ) then

        read(val,*) lctable

     else if( key == 'LCFOURIER' ) then

        read(val,*) lcfourier

     else if( key == 'LCMAG' ) then

        read(val,*) lcmag

     else if( key == 'LCAMP' ) then

        read(val,*) mamp

     else if( key == 'LCJD0' ) then

        read(val,*) lcjd0

     else if( key == 'LCPER' ) then

        read(val,*) lcper

     else if( key == 'LCCOO' ) then

        read(val,*) lcoo
        lccoo_init = .true.

     else if( key == 'NFILES' ) then

        read(val,*) nfiles

     else if( key == 'OUTPUT' ) then

        read(val,*) filemask

     end if

  end do

20 continue

  crpix = naxes / 2

  if( spread == 'AUTO' ) then
     if( exptime < 3 ) then
        spread = 'RANDOM'
     else
        spread = 'FFT'
     end if
  end if

  ! setup noise by Poisson
  if( bnoise < 0 .and. blevel > 0 ) bnoise = sqrt(blevel)

  ! whatever LC is generated
  if( lcmodel /= '' .or. lctable /= '' .or. lcfourier /= '' ) lcmode = .true.

  if( lcmode ) then
     if( .not. lccoo_init ) lcoo = crval

     if( lctable /= '' ) then
        call lctable_init(lctable)
     end if
     if( lcfourier /= '' ) then
        call lcfourio(lcfourier,fcoeff)
     end if

     lcamp = 10**(0.4*mamp)
  end if

  ! stars
  if( cat /= '' ) then
     call catstars(cat,labels,scale,angle,crval,crpix,phsystable,phsystem, &
          filter,xcoo,ycoo,flux)
  else
     call genstars(maglim,naxes,xcoo,ycoo,flux)
  end if

  if( verbose ) then
     write(error_unit,*)
     write(error_unit,*) "LIST OF PARAMETERS"
     write(error_unit,*)
     write(error_unit,*) "Point spread function (PSF): ",trim(psf)
     write(error_unit,*) "Seeing spread method: ",trim(spread)
     write(error_unit,*) "Atmosphere modelling: ",atmosphere
     write(error_unit,'(a,f0.2)') " HWHM = ",hwhm
     if( psf == 'MOFFAT' ) then
        write(error_unit,'(a,f0.2)') " Moffat exponent (beta) = ",beta
     else if( psf == 'SEEING' ) then
        write(error_unit,'(a,f0.2)') " Airy disk radius (airy) = ",airy
     end if
     write(error_unit,'(a,f0.2)') " Quantum efficiency = ",qefficiency
     if( atmosphere ) then
        write(error_unit,'(a,f0.2)') " Extinction coefficient = ",extink
     end if
     write(error_unit,*) "Photometry system: ",trim(phsystem)
     write(error_unit,*) "Filter: ",trim(filter)
     write(error_unit,'(a,f0.1)') " Background level = ",blevel
     write(error_unit,'(a,f0.1)') " Background noise = ",bnoise
     write(error_unit,'(a,2(f0.1,2x))') " Background gradient = ",bgrad
     write(error_unit,'(a,f0.2,a)') " Area = ",area," [m2]"
     write(error_unit,'(a,f0.1,a)') " Exposure time = ",exptime," [s]"
     write(error_unit,'(a,f0.1,a)') " Time span = ",timestep," [s]"
     write(error_unit,'(a,i4.4,2("-",i2.2))') " Exposure start date: ", &
          nint(year),nint(month),nint(day)
     write(error_unit,'(a,2(i2.2,":"),f0.3)') " Exposure start time: ", &
          nint(hour),nint(min),sec
     write(error_unit,'(a,i0," x ",i0)') " Width and height of image(s): ",naxes
     write(error_unit,'(a,2f7.2,a)') " Equatorial coordinates: ",crval," [deg]"
     write(error_unit,'(a,f0.1,a,g0.1,a)') ' Scale: ',1/scale," [pix/deg],  ", &
          scale*3600," [arcsec/pix]"
     write(error_unit,'(a,f0.1,a)') ' Field rotation = ',angle," [deg]"
     write(error_unit,'(a,f0.1,a)') ' Field of view = ',fov," [deg]"
     if( cat == '' ) then
        write(error_unit,*) "Star coordinates and fluxes: random generated"
        write(error_unit,'(a,i0)') " Magnitude limit = ",maglim
     else
        write(error_unit,*) "Star coordinates and fluxes: by `",trim(cat),"'"
     end if
     write(error_unit,'(a,2f7.2,a)') " Observatory coordinates: ", &
          longitude, latitude," [deg]"
     write(error_unit,*) "Light curve modelling: ",lcmode
     if( lcmode ) then
        if( lcmodel /= '' ) then
           write(error_unit,*) "Light curve model: ",trim(lcmodel)
        else if( lctable /= '' ) then
           write(error_unit,*) "Light curve in table: `",trim(lctable),"'"
        else if( lcfourier /= '' ) then
           write(error_unit,*) "Light curve by Fourier coefficients in: `", &
                trim(lcfourier),"':"
           do i = 0,size(fcoeff)-1
              write(error_unit,'(a,i0,a,2(3x,f0.5),a)') "   c(",i,") = ",fcoeff(i),"i"
           end do
        end if
        write(error_unit,'(a,f0.1)') " Light curve mean magnitude = ",lcmag
        write(error_unit,'(a,f0.1)') " Light curve amplitude = ",lcamp
        write(error_unit,'(a,f0.6)') " Light curve JD0 = ",lcjd0
        write(error_unit,'(a,f0.6,a)') " Light curve period = ",lcper," [day(s)]"
        write(error_unit,'(a,2f7.2,a)') " Coordinates of the object: ",lcoo," [deg]"
     end if
     write(error_unit,*)
     write(error_unit,*) "Filename, JD, airmass, seeing radius:"
  end if


  ! determine '?' in filemask
  n = 0
  do i = 1,len(filemask)
     if( filemask(i:i) == '?' ) n = n + 1
  end do
  if( nfiles > 1 .and. n == 0 ) n = int(log10(real(nfiles)) + 1)
  if( nfiles == 0 .and. n == 0 ) file = filemask
  write(fmt,'(a,i0,a)') '(a,i0.',n,',a)'

  if( datetime_init ) jd = datjd(year,month,day + (hour + (min + sec/60)/60)/24)
  timestep = timestep / 86400


  if( nfiles == 0 ) then

     call artfits(file,jd)

  else

     l1 = index(filemask,'?')

     if( l1 > 0 ) then
        l1 = l1 - 1
        l2 = index(filemask,'?',back=.true.) + 1

        do i = 1,nfiles
           write(file,fmt) filemask(1:l1),i,filemask(l2:len_trim(filemask))
           call artfits(file,jd)
           jd = jd + timestep
        end do
     else
        ! single filename
        call artfits(filemask,jd)
     end if

  end if

  if( allocated(fcoeff) ) deallocate(fcoeff)
  if( allocated(xcoo) ) deallocate(xcoo,ycoo,flux)
  call lcurve_finish

  stop 0

contains

  subroutine artfits(filename,jd)

    use generator
    use marker
    use spray
    use astrosphere
    use lcurves

    real(dbl), parameter :: rad = 57.295779513082322865_dbl

    character(len=*), intent(in) :: filename
    real(dbl), intent(in) :: jd

    integer :: status,blocksize
    character(len=FLEN_COMMENT) :: com
    character(len=FLEN_VALUE) :: datetime
    real(dbl), dimension(:,:), allocatable :: sky
    real(dbl) :: c,s,x,qe,see,extin,hwhms,jd2, ms, lcflux, phi
    integer :: year, month, day, hour, min, sec, m
    integer, dimension(2) :: maxes
    real(dbl), dimension(:), allocatable :: xstar,ystar,fluxes,xvar, yvar, fvar
    type(SprayType) :: spsf

    if( atmosphere ) then

       ! compute airmass
       x = xairmass(jd,longitude,latitude,crval(1),crval(2))
       if( x < 0 ) then
          call jdatetime(jd,year,month,day,hour,min,sec,ms)
          write(error_unit,'(a,i4.4,2(a,i2.2),a,2(i2.2,a),i2.2,a,i0.3,a)') &
               "Airmass undefined (below horizon) for date  ", &
               year,"-",month,"-",day,"  and time ",hour,":",min,":",sec,".", &
               nint(ms)," UT"
          return
       end if

       ! simulate seeing and extinction
       extin = exp(-extink*(x-1))          ! relative extinction against X=1
       qe = extin * qefficiency

       ! http://www.astro.auth.gr/~seeing-gr/seeing_gr_files/theory/node17.html
       ! Fries parameter
       see = x**0.6
       hwhms = hwhm*see

       if( verbose ) &
            write(error_unit,'(a,f15.5,f8.3,2f6.2)') trim(filename),jd,x,extin,see
    else
       hwhms = hwhm
       qe = qefficiency

       if( verbose ) &
            write(error_unit,'(a,f15.5)') trim(filename),jd
    endif

    if( lcmode ) then
       ! variables

       phi = phase(jd,lcjd0,lcper)

       if( lcmodel == 'WAVE' ) then
          lcflux = lcamp*wave(phi)
       else if( lcmodel == 'CEPHEID' ) then
          lcflux = lcamp*delta_cep(phi)
       else if( lcmodel == 'ALGOL' ) then
          lcflux = lcamp*beta_per(phi)
       else if( lcmodel /= '' ) then
          stop "Unrecognised light curve model. Try `Cepheid', `Algol' or `wave'."
       end if

       if( lcfourier /= '' ) then
          lcflux = lcamp*fourfun(fcoeff,phi)
       end if

       if( lctable /= '' ) then
          lcflux = lcspline(phi)
       end if

       allocate(xvar(1),yvar(1),fvar(1))
       call rain((/lcoo(1)/),(/lcoo(2)/),(/lcmag/),scale,angle,crval,crpix, &
            phsystable,phsystem,filter,xvar,yvar,fvar)

       fvar = fvar * lcflux
    else
       allocate(xvar(0),yvar(0),fvar(0))
    end if

    allocate(sky(naxes(1),naxes(2)))
    sky = 0

    ! background
    call background(blevel,bnoise,bgrad,sky)

    ! merge both lists: variables and ordinary stars
    m = size(xvar)
    n = m + size(xcoo)
    allocate(xstar(n),ystar(n),fluxes(n))
    xstar(1:m) = xvar;  xstar(m+1:) = xcoo
    ystar(1:m) = yvar;  ystar(m+1:) = ycoo
    fluxes(1:m) = fvar; fluxes(m+1:) = flux

    ! update fluxes
    fluxes = fluxes * exptime * area * qe

    ! simulate profile
    call spray_init(psf,spread,hwhms,airy,beta,spsf)

    ! create the artificial sky
    call spreadstars(spsf,xstar,ystar,fluxes,sky)

    ! watermark
    call watermark(sky,5*bnoise)

    ! cleanup
    if( allocated(xvar) ) deallocate(xvar,yvar,fvar)
    deallocate(xstar,ystar,fluxes)

    ! save the artificial sky
    call fitsio_unlink(filename)
    status = 0
    call ftinit(15,filename,blocksize,status)
    if( status /= 0 ) then
       write(error_unit,*) "Error: `",trim(filename),"' failed to create."
       goto 666
    end if

    ! convert JD to date
    call jdatetime(jd,year,month,day,hour,min,sec,ms)
    jd2 = jd + exptime / 2.0_dbl / 86400.0_dbl

    ! extension with artifical frame
    call ftphps(15,bitpix,size(naxes),naxes,status)
    write(datetime,"(i4.4,'-',i2.2,'-',i2.2,'T',i2.2,':',i2.2,':',i2.2,'.',i3.3)") &
         year, month, day, hour, min, sec, nint(ms)
    call ftpkys(15,FITS_KEY_DATEOBS,datetime,'date and time in UT',status)
    call ftpkys(15,FITS_KEY_FILTER,filter,'filter',status)
    call ftpkyd(15,FITS_KEY_JD,jd2,-16,'[day] Julian date at half of exposure',status)
    call ftpkyd(15,FITS_KEY_EXPTIME,exptime,-3,'[s] exposure time',status)
    call ftpkyd(15,FITS_KEY_AREA,area,-2,'[m2] detection area',status)
    call ftpkyd(15,FITS_KEY_HWHM,hwhms,-3,'[pix] half with at half maximum',status)
    if( atmosphere ) &
         call ftpkyd(15,FITS_KEY_AIRMASS,x,-4,'airmass',status)
    call ftpkys(15,'PSF',psf,'PSF',status)
    if( psf == 'MOFFAT' ) then
       call ftpkyd(15,'MOFFATB',beta,-3,'Moffat exponent',status)
    else if ( psf == 'SEEING' ) then
       call ftpkyd(15,'AIRYDISK',airy,-3,'[pix] Airy disk radius',status)
       call ftpkys(15,'SPREAD',spread,'seeing spread method ',status)
    end if
    call ftpkys(15,'CTYPE1','RA---TAN','the coordinate type for the first axis',status)
    call ftpkys(15,'CTYPE2','DEC--TAN','the coordinate type for the first axis',status)
    call ftpkye(15,'CRPIX1',real(naxes(1)/2),-6,'[pix]',status)
    call ftpkye(15,'CRPIX2',real(naxes(2)/2),-6,'[pix]',status)
    call ftpkyd(15,'CRVAL1',crval(1),-17,'[deg]',status)
    call ftpkyd(15,'CRVAL2',crval(2),-17,'[deg]',status)
    s = sin(angle/rad)
    c = cos(angle/rad)
    call ftpkyd(15,'CD1_1',-c*scale,13,'',status)
    call ftpkyd(15,'CD1_2',s*scale,13,'',status)
    call ftpkyd(15,'CD2_1',s*scale,13,'',status)
    call ftpkyd(15,'CD2_2',c*scale,13,'',status)
    call ftpkyd(15,FITS_KEY_LONGITUDE,longitude,-5,'[deg] geographic longitude (-east)',status)
    call ftpkyd(15,FITS_KEY_LATITUDE,latitude,-5,'[deg] geographic latitude (+north)',status)
    call ftpcom(15,'Artificial sky by Munipack.',status)
    call ftpcom(15,'Parameters:',status)
    write(com,'(a,4(g0.6,1x))') 'Background: level, noise, gradient = ',&
         blevel,bnoise,bgrad
    call ftpcom(15,com,status)
    write(com,'(2a)') 'PSF = ',trim(psf)
    call ftpcom(15,com,status)
    write(com,'(2a)') 'Seeing spread method = ',trim(spread)
    call ftpcom(15,com,status)
    write(com,'(a,i0)') 'Magnitude limit  = ',maglim
    call ftpcom(15,com,status)
    write(com,'(a,g0.3)') 'Quantum efficiency  = ', qe
    call ftpcom(15,com,status)
    if( atmosphere ) then
       write(com,'(a,g0.3)') 'Extinction coefficient  = ', extink
       call ftpcom(15,com,status)
       write(com,'(a,g0.3)') 'Seeing factor  = ', see
       call ftpcom(15,com,status)
    end if
    call ftpcom(15,MUNIPACK_VERSION,status)
    call ftp2dd(15,1,naxes(1),naxes(1),naxes(2),sky,status)

    ! PSF extension
    maxes = (/ size(spsf%psf,1),size(spsf%psf,2) /)
    call ftiimg(15,-32,2,maxes,status)
    call ftukys(15,'EXTNAME','PSF','',status)
    call ftpkyj(15,'ZOOM',spsf%zoom,'zoom',status)
    call ftp2dd(15,1,maxes(1),maxes(1),maxes(2),spsf%psf,status)

666 continue

    call ftclos(15,status)
    call ftrprt('STDERR',status)

    deallocate(sky)
    call spray_clear(spsf)

  end subroutine artfits


  subroutine lctable_init(tablename)

    use lcio

    character(len=*), intent(in) :: tablename
    real(dbl), dimension(:), allocatable :: time, flux, dflux
    real(dbl), parameter :: sopt = -1

    call lcimport(tablename,time,flux,dflux)

    call lcurves_init(time,flux,dflux,sopt)
    deallocate(time,flux,dflux)

  end subroutine lctable_init


end program artificial
