!
!  FITS I/O for photometric calibration
!
!  Copyright © 2012-6 F.Hroch (hroch@physics.muni.cz)
!
!  This file is part of Munipack.
!
!  Munipack is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  Munipack is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with Munipack.  If not, see <http://www.gnu.org/licenses/>.

module mfits

  use iso_fortran_env
  use fitsio

  implicit none

  integer, parameter, private :: dbl = selected_real_kind(15)
  real(dbl), parameter, private :: rad = 57.295779513082322865_dbl
  logical, private :: verbose = .true.


contains


  subroutine readcat(cat,labels,label_mag,label_magerr, &
       alpha,delta,mag,magerr,catid,status)

    character(len=*), intent(in) :: cat
    character(len=*), dimension(:), intent(in) :: labels,label_mag,label_magerr
    real(dbl), dimension(:), allocatable, intent(out) :: alpha,delta
    real(dbl), dimension(:,:), allocatable, intent(out) :: mag,magerr
    character(len=*), intent(out) :: catid
    integer, intent(out) :: status

    integer, parameter :: frow = 1, felem = 1
    real(dbl), parameter :: nullval = 99.99999

    integer :: nrows, ncols, i
    integer, dimension(size(labels)) :: cols
    integer, dimension(size(label_mag)) :: col_mag
    integer, dimension(size(label_magerr)) :: col_magerr
    character(len=FLEN_CARD) :: buf
    logical :: anyf


    status = 0

    ! open and move to a table extension
    call fttopn(15,cat,0,status)
    call ftgnrw(15,nrows,status)
    if( status /= 0 ) goto 666

    ! define reference frame and identification of catalogue
    call ftgkys(15,'EXTNAME',catid,buf,status)
    if( status == KEYWORD_NOT_FOUND ) then
       catid = ''
       status = 0
    end if

    ! find columns by labels
    do i = 1, size(labels)
       call ftgcno(15,.true.,labels(i),cols(i),status)
    end do
    do i = 1, size(label_mag)
       call ftgcno(15,.true.,label_mag(i),col_mag(i),status)
    end do
    do i = 1, size(label_magerr)
       call ftgcno(15,.true.,label_magerr(i),col_magerr(i),status)
    end do
    if( status /= 0 ) goto 666

    if( nrows > 0 ) then
       ncols = size(label_mag)
       allocate(alpha(nrows),delta(nrows),mag(nrows,ncols),magerr(nrows,ncols))

       magerr = 9.99999

       call ftgcvd(15,cols(1),frow,felem,nrows,nullval,alpha,anyf,status)
       call ftgcvd(15,cols(2),frow,felem,nrows,nullval,delta,anyf,status)

       do i = 1, size(col_mag)
          call ftgcvd(15,col_mag(i),frow,felem,nrows,nullval,mag(:,i),anyf,status)
       end do
       do i = 1, size(col_magerr)
          call ftgcvd(15,col_magerr(i),frow,felem,nrows,nullval/10,magerr(:,i),anyf,status)
       end do
    end if

    call ftclos(15,status)
    if( status /= 0 ) goto 666

    return

666 continue

    call ftclos(15,status)
    call ftrprt('STDERR',status)

    if( allocated(mag) ) deallocate(mag,magerr,alpha,delta)

  end subroutine readcat


  subroutine readframe(filename,keys,alpha,delta,cts,dcts,ftol,exptime,area, &
       photosys,filter,init_area,saper,status,amass)

    use astrotrafo
    use astrosphere
    use trajd
    use phio

    character(len=*), intent(in) :: filename
    character(len=*), dimension(:), intent(in) :: keys
    real(dbl), dimension(:), allocatable, intent(out) :: alpha,delta,cts,dcts
    real(dbl), intent(out) :: ftol,exptime, area
    character(len=*), intent(out) :: photosys,filter
    logical, intent(in) :: init_area
    integer, intent(in) :: saper
    integer, intent(out) :: status
    real(dbl), optional, intent(out) :: amass

    integer, parameter :: frow = 1
    integer, parameter :: felem = 1
    integer, parameter :: DIM = 2
    integer, dimension(DIM) :: naxes
    integer :: naxis, nrows, colnum, naper, maper, s1, s2, s4, s5, yy,m,d,hour,min, i
    real(dbl) :: year, month, day, sec, jd, longitude, latitude
    character(len=FLEN_COMMENT) :: com
    character(len=FLEN_CARD) :: key, keycts, keyerr
    real(dbl), dimension(:), allocatable :: x,y
    real(dbl) :: nullval
    real :: hwhm, aper
    logical :: anyf
    logical :: undef_airmass, have_grow, have_ap
    type(AstroTrafoProj) :: tproj


    ! input FITS file
    status = 0
    call ftiopn(20,filename,0,status)
    call ftgidm(20,naxis,status)
    call ftgisz(20,DIM,naxes,status)
    if( status /= 0 ) goto 666

    ! read astrometric calibration
    call wcsget(20,tproj,status)
    if( status == KEYWORD_NOT_FOUND ) then
       write(error_unit,*) 'Astrometry keywords not found in header.'
       goto 666
    end if
    ftol = 5.0*tproj%err

    ! Citizen date
    call datetime(20,(/keys(3),keys(8)/),yy,m,d,hour,min,sec,status)
    if( status == 0 ) then
       year = yy
       month = m
       day = d + (hour + (min + sec/60.0_dbl)/60.0_dbl)/24.0_dbl
       jd = datjd(year,month,day)
    else
       write(error_unit,*) "Warning: Failed to get date or time."
       goto 666
    end if

    ! Geographic coordinates
    call ftgkyd(20,keys(6),longitude,com,status)
    call ftgkyd(20,keys(7),latitude,com,status)
    if( status == KEYWORD_NOT_FOUND ) then
       write(error_unit,*) &
            "Warning: Geographical longitude or latitude identified by FITS keywords `",&
            trim(keys(6)),"',`",trim(keys(7)),"' not found. Airmass is undefined."
       status = 0
       undef_airmass = .true.
    else
       undef_airmass = .false.
    end if

    s1 = 0
    call ftgkyd(20,keys(1),exptime,com,s1)
    if( s1 == KEYWORD_NOT_FOUND ) then
       write(error_unit,*) "Warning: Exposure time identified by FITS keyword `",&
            trim(keys(1)),"' not found. Set to default 1 sec."
       exptime = 1
       s1 = 0
    end if

    s2 = 0
    if( .not. init_area ) then
       call ftgkyd(20,keys(2),area,com,s2)
       if( s2 == KEYWORD_NOT_FOUND ) then
          write(error_unit,*) "Warning: Area identified by FITS keyword `", &
               trim(keys(2)),"' not found. Set to default 1 sq. meter."
          area = 1
          s2 = 0
       end if
    end if

!    s3 = 0
!    if( present(airmass) ) then
!       call ftgkyd(20,keys(3),airmass,com,s3)
!       if( s3 == KEYWORD_NOT_FOUND ) then
!          write(error_unit,*) "Warning: Airmass identified by FITS keyword `", &
!               trim(keys(3)),"' not found. Default is 0 (extra-atmospheric)."
!          airmass = 0
!          s3 = 0
!       end if
!    end if

    s4 = 0
    call ftgkys(20,keys(4),photosys,com,s4)
    if( s4 == KEYWORD_NOT_FOUND ) then
!       write(error_unit,*) "Photometry system by FITS keyword `", &
!            trim(keys(4)),"' not found, leaving it empty."
       photosys = ''
       s4 = 0
    end if

    s5 = 0
    call ftgkys(20,keys(5),filter,com,s5)
    if( s5 == KEYWORD_NOT_FOUND ) then
       write(error_unit,*) "Filter identified by FITS keyword `", &
            trim(keys(5)),"' not found. Set to default `' (empty)."
       filter = ''
       s5 = 0
    end if

    status = s1 + s2 + s4 + s5
    if( status /= 0 ) goto 666

    ! Julian date at midpoint of exposure
    jd = jd + exptime / 2.0_dbl / 86400.0_dbl

    if( present(amass) ) then
       if( .not. undef_airmass ) then
          ! airmass at the time
          amass = xairmass(jd,longitude,latitude,tproj%acen,tproj%dcen)
       else
          amass = -1
       end if
    end if

    ! detect all possible photometry tables
!    phexts(1) = APEREXTNAME
!    phexts(2) = GROWEXTNAME
!    have_phext = .false.
!    do i = 1,size(phexts)
!       call ftmnhd(20,BINARY_TBL,phexts(i),0,status)
!       if( status == 0 ) have_phext(i) = .true.
!       if( status == BAD_HDU_NUM ) status = 0
!       if( status /= 0 ) goto 666
!    end do



    ! try to select grow-curve photometry table
    call ftmnhd(20,BINARY_TBL,GROWEXTNAME,0,status)
    have_grow = .not. (status == BAD_HDU_NUM)

    ! try to select extension with aperture photometry
    if( .not. have_grow ) then
       status = 0
       call ftmnhd(20,BINARY_TBL,APEREXTNAME,0,status)
       have_ap = .not. (status == BAD_HDU_NUM)
    else
       have_ap = .true.
    end if

    if( .not. (have_grow .or. have_ap) )then
       write(error_unit,*) &
            "Error: Failed to find an aperture or a growth photometry extension: ", &
            trim(filename)
       goto 666
    end if

    ! key for HWHM is presented in all photometry extensions
    ! (methods for their estimates are different)
    call ftgkye(20,FITS_KEY_HWHM,hwhm,com,status)
    ftol = hwhm * tproj%scale

    keycts = ''
    keyerr = ''

    if( have_grow .and. saper == 0 ) then
       keycts = FITS_COL_GCOUNT
       keyerr = FITS_COL_GCOUNTERR
    end if

    if( (.not. have_grow .and. have_ap) .or. (have_ap .and. saper > 0) ) then

       call ftmnhd(20,BINARY_TBL,APEREXTNAME,0,status)

       if( saper > 0 ) then
          maper = saper
       else
          call ftgkyj(20,FITS_KEY_NAPER,naper,com,status)
          do i = 1, naper
             call ftkeyn(FITS_KEY_APER,i,key,status)
             call ftgkye(20,key,aper,com,status)
             if( aper > 3*hwhm ) then
                maper = i
                exit
             end if
          end do
       end if

       write(keycts,'(a,i0)') FITS_COL_APCOUNT,maper
       write(keyerr,'(a,i0)') FITS_COL_APCOUNTERR,maper

    end if

    call ftgnrw(20,nrows,status)
    if( status /= 0 ) goto 666

    allocate(x(nrows),y(nrows))

    ! table
    call ftgcno(20,.true.,FITS_COL_X,colnum,status)
    call ftgcvd(20,colnum,frow,felem,nrows,nullval,x,anyf,status)
    call ftgcno(20,.true.,FITS_COL_Y,colnum,status)
    call ftgcvd(20,colnum,frow,felem,nrows,nullval,y,anyf,status)

    if( status == 0 ) then
       allocate(alpha(nrows), delta(nrows))
       call invtrafo(tproj,x,y,alpha,delta)
    end if
    deallocate(x,y)

    allocate(cts(nrows),dcts(nrows))
    call ftgcno(20,.true.,keycts,colnum,status)
    call ftgcvd(20,colnum,frow,felem,nrows,nullval,cts,anyf,status)
    call ftgcno(20,.true.,keyerr,colnum,status)
    call ftgcvd(20,colnum,frow,felem,nrows,nullval,dcts,anyf,status)

666 continue

    call ftclos(20,status)
    call ftrprt('STDERR',status)

  end subroutine readframe

  subroutine checkframe(filename,keys,filter,nrows,status)

    character(len=*), intent(in) :: filename
    character(len=*), dimension(:), intent(in) :: keys
    character(len=*), intent(out) :: filter
    integer, intent(out) :: nrows
    integer, intent(in out) :: status

    character(len=FLEN_COMMENT) :: com

    call ftiopn(20,filename,0,status)
    call ftgkys(20,keys(5),filter,com,status)
    if( status == KEYWORD_NOT_FOUND ) then
       write(error_unit,*) "Filter identified by FITS keyword `", &
            trim(keys(5)),"' not found. Default is `' (empty)."
       filter = ''
       status = 0
    end if

    call ftmnhd(20,BINARY_TBL,APEREXTNAME,0,status)
    if( status == BAD_HDU_NUM ) then
       if( verbose ) write(error_unit,*) "Failed to find a photometry extension: ",&
            trim(filename)
       goto 666
    end if

    call ftgnrw(20,nrows,status)

666 continue

    call ftclos(20,status)
    call ftrprt('STDERR',status)

  end subroutine checkframe


  subroutine readframes(filenames,keys,filters_ref,photosys,ra,dec,cts,dcts,utol,&
            exptime,area,filter,tol,init_area,saper,status)

    character(len=*), dimension(:), intent(in) :: filenames,keys,filters_ref
    character(len=*), intent(out) :: photosys
    real(dbl), dimension(:,:), allocatable, intent(out) :: ra,dec,cts,dcts
    real(dbl), dimension(:), allocatable, intent(out) :: exptime,area,tol
    character(len=FLEN_VALUE), dimension(:), allocatable, intent(out) :: filter
    real(dbl), intent(in) :: utol
    integer, intent(in) :: saper
    logical, intent(in) :: init_area
    integer, intent(out) :: status

    real(dbl) :: ftol
    integer :: n, nd, nfiles!, i, j
    real(dbl), dimension(:), allocatable :: qra,qdec,dn,ddn
!    integer, allocatable, dimension(:) :: ndat, order
    integer, allocatable, dimension(:) :: order
    character(len=FLEN_VALUE) :: phsys

    nfiles = size(filenames)

    call file_order(filenames,keys,filters_ref,order,nd)

!!$    allocate(filter(nfiles),ndat(nfiles),order(nfiles))
!!$
!!$    ! before loading of the (big) data, we are discovering their sizes and filters
!!$    status = 0
!!$    do n = 1, nfiles
!!$       call checkframe(filenames(n),keys,filter(n),ndat(n),status)
!!$       if( status /= 0 ) then
!!$          write(error_unit,*) "File: `",trim(filenames(n)),"'"
!!$          stop 'Failed to read a frame.'
!!$       end if
!!$    end do
!!$
!!$    ! array allocation dimension
!!$    nd = maxval(ndat)
!!$
!!$    ! and using the info to arrange order by filters
!!$    n = 0
!!$    do i = 1,size(filters_ref)
!!$       do j = 1,size(filter)
!!$          if( filters_ref(i) == filter(j) ) then
!!$             n = n + 1
!!$             order(n) = j
!!$          end if
!!$       end do
!!$    end do
!!$    if( n /= nfiles ) &
!!$         stop 'Filters of frames does not unique corresponds to ones of standard set.'

    ! now, we are prepared to read the data
    allocate(exptime(nfiles),area(nfiles),tol(nfiles),filter(nfiles), &
         cts(nd,nfiles),dcts(nd,nfiles),ra(nd,nfiles),dec(nd,nfiles))

    cts = -1
    dcts = -1
    ra = 0
    dec = 0

    status = 0
    do n = 1, nfiles

       call readframe(filenames(order(n)),keys,qra,qdec,dn,ddn,ftol,&
            exptime(n),area(n),phsys,filter(n),init_area,&
            saper,status)

       if( status /= 0 ) then
          write(error_unit,*) "File: `",trim(filenames(n)),"'"
          stop 'Failed to read a frame.'
       end if

       if( utol > epsilon(utol) ) then
          tol(n) = utol
       else
          tol(n) = ftol
       end if

       if( n == 1 ) then
          photosys = phsys
       else
          if( phsys /= photosys ) write(error_unit,*) &
               "Uncompatible photometry systems: `",trim(phsys),"' and `", &
               trim(photosys),"'."
       end if

       nd = size(dn)

       cts(1:nd,n) = dn
       dcts(1:nd,n) = ddn
       ra(1:nd,n) = qra
       dec(1:nd,n) = qdec

       deallocate(qra,qdec,dn,ddn)

    end do
    deallocate(order)

  end subroutine readframes


  subroutine file_order(filenames,keys,filters_ref,order,maxdat)

    character(len=*), dimension(:), intent(in) :: filenames,keys,filters_ref
    integer, dimension(:), allocatable, intent(out) :: order
    integer, intent(out), optional :: maxdat
    character(len=FLEN_VALUE), dimension(:), allocatable :: filter
    integer, dimension(:), allocatable :: ndat
    integer :: i,j,n,status,nfiles,nfilters

    nfilters = size(filters_ref)
    nfiles = size(filenames)
    allocate(filter(nfiles),ndat(nfiles),order(nfiles))

    ! before loading of the (big) data, we are discovering their sizes and filters
    status = 0
    do n = 1, nfiles
       call checkframe(filenames(n),keys,filter(n),ndat(n),status)
       if( status /= 0 ) then
          write(error_unit,*) "File: `",trim(filenames(n)),"'"
          stop 'Failed to read a frame.'
       end if
    end do

    ! and using the info to arrange order by filters
    n = 0
    do i = 1,nfilters
       do j = 1,size(filter)
          if( filters_ref(i) == filter(j) ) then
             n = n + 1
             order(n) = j
          end if
       end do
    end do

    if( n /= nfiles ) then
       write(error_unit,*)
       write(error_unit,*) 'Summary of filter survey: '
       write(error_unit,*) 'Reference filters: ',(trim(filters_ref(i)),i=1,nfilters)
       write(error_unit,*) 'List of files and filters:'
       do i = 1, nfiles
          write(error_unit,*) trim(filenames(i)),': `',trim(filter(i)),"'"
       end do
       stop 'Filters of frames does not corresponds to the standard filter set.'
    end if

    if( present(maxdat) ) then
       ! array allocation dimension
       maxdat = maxval(ndat)
    end if

    deallocate(filter,ndat)

  end subroutine file_order

end module mfits
